@@ -7,33 +7,13 @@ Functions having to do with parser keywords, like testing if a function is a blo
 #define FISH_PARSER_KEYWORD_H
 
 /**
-   Return valuse for parser_keywords_is_switch()
-*/
-enum 
-{
-	ARG_NON_SWITCH,
-	ARG_SWITCH,
-	ARG_SKIP
-}
-	;
-
-
-
-/**
-   Check if the specified argument is a switch. Return ARG_SWITCH if yes,
-   ARG_NON_SWITCH if no and ARG_SKIP if the argument is '--'
-*/
-int parser_keywords_is_switch( const wchar_t *cmd );
-
-
-/**
-   Tests if the specified commands parameters should be interpreted as another command, which will be true if the command is either 'command', 'exec', 'if', 'while' or 'builtin'.  
+   Tests if the specified commands parameters should be interpreted as another command, which will be true if the command is either 'command', 'exec', 'if', 'while', or 'builtin'.  This does not handle "else if" which is more complicated.
 
    \param cmd The command name to test
    \return 1 of the command parameter is a command, 0 otherwise
 */
 
-int parser_keywords_is_subcommand( const wchar_t *cmd );
+bool parser_keywords_is_subcommand(const wcstring &cmd);
 
 /**
    Tests if the specified command is a reserved word, i.e. if it is
@@ -44,20 +24,20 @@ int parser_keywords_is_subcommand( const wchar_t *cmd );
    \param word The command name to test
    \return 1 of the command parameter is a command, 0 otherwise
 */
-int parser_keywords_is_reserved( const wchar_t *word );
+bool parser_keywords_is_reserved(const wcstring &word);
 
 /**
    Test if the specified string is command that opens a new block
 */
 
-int parser_keywords_is_block( const wchar_t *word);
+bool parser_keywords_is_block(const wcstring &word);
 
 /**
    Check if the specified command is one of the builtins that cannot
    have arguments, any followin argument is interpreted as a new
    command
 */
-int parser_keywords_skip_arguments( const wchar_t *cmd );
+bool parser_keywords_skip_arguments(const wcstring &cmd);
 
 
 #endif
